package filegateway

import (
	"context"
	"crypto/tls"
	"crypto/x509"
	"encoding/pem"
	"errors"
	"fmt"
	"io"
	"net"
	"os"
	"path"
	"strings"
	"time"

	"github.com/rs/zerolog"
	"github.com/secsy/goftp"
	xbus "xbus.io/go-xbus/v4"
)

// FTPFlavor is the type of the remote FTP server
type FTPFlavor int

const (
	// FTPFlavorDefault is for modern ftp servers that implements the MLSD command
	FTPFlavorDefault FTPFlavor = iota
	// FTPFlavorVSFTP is for a vsftp server
	FTPFlavorVSFTP
)

var (
	ErrMaxAttemptReached = errors.New("max attempts reached")
	ErrNoADirectory      = errors.New("not a directory")
)

func loadServerCert(cert string) (*x509.Certificate, error) {
	pemBlock := []byte(cert)

	var derBlock *pem.Block
	for {
		derBlock, pemBlock = pem.Decode(pemBlock)
		if derBlock == nil {
			break
		}
		if derBlock.Type == "CERTIFICATE" {
			break
		}
		derBlock = nil
	}
	if derBlock == nil {
		return nil, fmt.Errorf("Found no certificate")
	}
	return x509.ParseCertificate(derBlock.Bytes)
}

func ftpFilesystemFactory(options FilesystemOptions, log zerolog.Logger) (Filesystem, error) {
	if options.Host == "" {
		return nil, fmt.Errorf(
			"Invalid host for a ftp(s):// uri. Expects a non-mpty host")
	}
	if options.Port == "" {
		options.Port = "21"
	}
	fs := FtpFilesystem{
		FilesystemOptions: options,
		address:           options.Host + ":" + options.Port,
		clientConfig: goftp.Config{
			User:     options.Username,
			Password: options.Password,
		},
		log: log,
	}
	if err := options.CustomStringD("basedir", &(fs.basedir), "/"); err != nil {
		return nil, err
	}
	if err := options.CustomStringD("tempdir", &(fs.tempdir), ""); err != nil {
		return nil, err
	}
	if err := options.CustomIntD("maxmembufsize", &(fs.maxMemBufSize), 1024*1024*16); err != nil {
		return nil, err
	}
	if err := options.CustomIntD("maxconn", &fs.clientConfig.ConnectionsPerHost, 4); err != nil {
		return nil, err
	}
	if err := options.CustomIntD("maxattempts", &fs.maxAttempts, 5); err != nil {
		return nil, err
	}

	if value, err := options.GetCustomStringD("flavor", "default"); err == nil {
		switch value {
		case "default":
			fs.flavor = FTPFlavorDefault
		case "vsftp":
			fallthrough
		case "vsftpd":
			fs.flavor = FTPFlavorVSFTP
		}
	} else {
		return nil, err
	}

	if options.Scheme == "ftps" {
		fs.clientConfig.TLSConfig = &tls.Config{
			ServerName: options.Host,
		}

		if tlsmode, err := options.GetCustomStringD("tls-mode", "explicit"); err == nil {
			switch tlsmode {
			case "explicit":
				fs.clientConfig.TLSMode = goftp.TLSExplicit
			case "implicit":
				fs.clientConfig.TLSMode = goftp.TLSImplicit
			default:
				return nil, fmt.Errorf("Invalid tls-mode. Expects \"explicit\" or \"implicit\", got %s", tlsmode)
			}
		} else {
			return nil, err
		}

		if servername, err := options.GetCustomStringD("tls-servername", ""); err == nil {
			if servername != "" {
				fs.clientConfig.TLSConfig.ServerName = servername
			}
		} else {
			return nil, err
		}

		if hostcert, err := options.GetCustomStringD("host-cert", ""); err == nil {
			if hostcert == "" {
				return nil, fmt.Errorf("'host-cert' is empty")
			}
			cert, err := loadServerCert(hostcert)
			if err != nil {
				return nil, fmt.Errorf("Error parsing 'host-cert': %s", err)
			}
			certpool := x509.NewCertPool()
			certpool.AddCert(cert)

			fs.clientConfig.TLSConfig.RootCAs = certpool
		} else {
			return nil, err
		}
	}
	if err := fs.initClient(); err != nil {
		return nil, err
	}
	return &fs, nil
}

// FtpFilesystem stores files on ftp(s) server
type FtpFilesystem struct {
	FilesystemOptions
	address       string
	clientConfig  goftp.Config
	basedir       string
	flavor        FTPFlavor
	tempdir       string
	maxMemBufSize int
	maxAttempts   int

	client *goftp.Client
	log    zerolog.Logger
}

func ftpErrIsNotExist(err error) bool {
	var goftpError goftp.Error
	if errors.As(err, &goftpError) && goftpError.Code() == 550 {
		return true
	}
	return errors.Is(err, os.ErrNotExist)
}

func ftpConvertError(err error) error {
	if ftpErrIsNotExist(err) {
		return os.ErrNotExist
	}
	return err
}

func (fs *FtpFilesystem) initClient() error {
	client, err := goftp.DialConfig(fs.clientConfig, fs.address)
	if err != nil {
		return err
	}
	fs.client = client
	return nil
}

func isTemporary(err error) bool {
	var ftpErr goftp.Error
	if errors.As(err, &ftpErr) && ftpErr.Temporary() {
		return true
	}
	var netErr net.Error
	if errors.As(err, &netErr) && netErr.Timeout() {
		return true
	}
	if strings.HasSuffix(err.Error(), "connect: connection reset by peer") {
		return true
	}
	return false
}

func (fs *FtpFilesystem) ftpAttempt(op string, call func(client *goftp.Client) error) error {
	var err error
	for attempt := 0; attempt != fs.maxAttempts; attempt++ {
		err = call(fs.client)
		if err == nil {
			if attempt != 0 {
				fs.log.Warn().Str("operation", op).Int("attempt", attempt).Msg("call succeeded after some attempts")
			}
			return nil
		}

		if isTemporary(err) {
			fs.log.Warn().Err(err).Str("operation", op).Int("attempt", attempt).Msg("got a ftp temporary error, trying again")

			continue
		}

		return ftpConvertError(err)
	}
	return fmt.Errorf("%w: last error was: '%s'", ErrMaxAttemptReached, err)
}

func (fs *FtpFilesystem) ftpStat(path string) (os.FileInfo, error) {
	var info os.FileInfo

	if err := fs.ftpAttempt("Stat", func(client *goftp.Client) error {
		res, err := client.Stat(path)
		if err != nil {
			return err
		}
		info = res
		return nil
	}); err != nil {
		return nil, err
	}

	return info, nil
}

func (fs *FtpFilesystem) ftpReadDir(path string) ([]os.FileInfo, error) {
	var infoList []os.FileInfo

	if err := fs.ftpAttempt("ReadDir", func(client *goftp.Client) error {
		res, err := client.ReadDir(path)
		if err != nil {
			return err
		}
		infoList = res
		return nil
	}); err != nil {
		return nil, err
	}

	return infoList, nil
}

func (fs *FtpFilesystem) ftpMkdir(path string) (string, error) {
	var created string

	if err := fs.ftpAttempt("Mkdir", func(client *goftp.Client) error {
		res, err := client.Mkdir(path)
		if err != nil {
			return err
		}
		created = res
		return nil
	}); err != nil {
		return "", err
	}

	return created, nil
}

func (fs *FtpFilesystem) mkdirAll(dirpath string) error {
	if fs.flavor == FTPFlavorVSFTP {
		return fs.mkdirAllNoStat(dirpath)
	}
	return fs.mkdirAllStat(dirpath)
}

func (fs *FtpFilesystem) mkdirAllStat(dirpath string) error {
	dirinfo, err := fs.ftpStat(dirpath)

	switch {
	case err != nil:
		if ftpErrIsNotExist(err) {
			if err := fs.mkdirAllStat(path.Dir(dirpath)); err != nil {
				return err
			}
			if _, err := fs.ftpMkdir(dirpath); err != nil {
				return err
			}
		}
	case !dirinfo.IsDir():
		return fmt.Errorf("%w: %s", ErrNoADirectory, dirpath)
	}
	return nil
}

func (fs *FtpFilesystem) mkdirAllNoStat(dirpath string) error {
	parent, name := path.Split(dirpath)
	parent = strings.TrimRight(parent, "/")

	if parent != "" && parent != "." && parent != "~" {
		if err := fs.mkdirAllNoStat(parent); err != nil {
			return err
		}
	}

	if parent == "" {
		parent = "/"
	}

	// Because we cannot use Stat(), we need to list the parent directory
	list, err := fs.ftpReadDir(parent)
	if err != nil {
		return err
	}

	for _, info := range list {
		if info.Name() == name {
			if info.IsDir() {
				return nil
			}
			return fmt.Errorf("%w: %s", ErrNoADirectory, dirpath)
		}
	}

	if _, err := fs.ftpMkdir(dirpath); err != nil {
		return err
	}

	// Check that the directory was properly created
	list, err = fs.ftpReadDir(parent)
	if err != nil {
		return err
	}

	for _, info := range list {
		if info.Name() == name {
			if info.IsDir() {
				return nil
			}
			return fmt.Errorf("%w: %s", ErrNoADirectory, dirpath)
		}
	}

	return fmt.Errorf("Mkdir failed (no more detail available): %s", dirpath)
}

func newFtpFileWriter(fs *FtpFilesystem, filepath string) *ftpFileWriter {
	return &ftpFileWriter{
		fs:       fs,
		filepath: filepath,

		buf:    NewOnDiskBuffer(int64(fs.maxMemBufSize), fs.tempdir, "xbus-filegateway-ftp-write-*"),
		closed: false,
	}
}

type ftpFileWriter struct {
	fs       *FtpFilesystem
	filepath string
	buf      *OnDiskBuffer

	closed bool
}

func (w *ftpFileWriter) Write(data []byte) (int, error) {
	return w.buf.Write(data)
}

func (w *ftpFileWriter) Close() error {
	w.closed = true
	return w.fs.ftpAttempt("Store", func(client *goftp.Client) error {
		if _, err := w.buf.Seek(0, io.SeekStart); err != nil {
			return err
		}
		return client.Store(w.filepath, w.buf)
	})
}

// Write ...
func (fs *FtpFilesystem) Write(filepath string, size int) (io.WriteCloser, error) {
	filepath = path.Join(fs.basedir, filepath)

	if err := fs.mkdirAll(path.Dir(filepath)); err != nil {
		return nil, err
	}

	return newFtpFileWriter(fs, filepath), nil
}

// Read returns an opened File
func (fs *FtpFilesystem) Read(filepath string) (io.ReadCloser, error) {
	filepath = path.Join(fs.basedir, filepath)

	if fs.flavor != FTPFlavorVSFTP {
		info, err := fs.ftpStat(filepath)
		if err != nil {
			return nil, err
		}
		if info.IsDir() {
			return nil, fmt.Errorf("Expected a file, found a dir: %s", filepath)
		}
	}

	var buffer = NewOnDiskBuffer(int64(fs.maxMemBufSize), fs.tempdir, "xbus-filegateway-ftp-write-*")

	if err := fs.ftpAttempt("Retrieve", func(client *goftp.Client) error {
		if err := buffer.Reset(); err != nil {
			return err
		}
		return fs.client.Retrieve(filepath, buffer)
	}); err != nil {
		if fs.flavor == FTPFlavorVSFTP {
			// XXX This is not very reliable, and inspecting the goftp error
			// would bring little added reliability. A more vsftp-adapted
			// client may be the way to go.
			if err.Error() == "unexpected response: 550-Failed to open file." {
				err = os.ErrNotExist
			}
		}
		return nil, err
	}

	return io.NopCloser(buffer), nil
}

// Rename renames a file
func (fs *FtpFilesystem) Rename(oldpath, newpath string) error {
	oldpath = path.Join(fs.basedir, oldpath)
	newpath = path.Join(fs.basedir, newpath)
	return fs.ftpAttempt("Rename", func(client *goftp.Client) error {
		return client.Rename(oldpath, newpath)
	})
}

// Delete removes a file
func (fs *FtpFilesystem) Delete(filepath string) error {
	filepath = path.Join(fs.basedir, filepath)
	return fs.ftpAttempt("Delete", func(client *goftp.Client) error {
		return client.Delete(filepath)
	})
}

// Observe naïvely do a file list every pollingInterval
func (fs *FtpFilesystem) Observe(
	ctx context.Context, pattern string, paths chan<- string, log xbus.Logger,
) error {
	patterndir := path.Dir(path.Join(fs.basedir, pattern))

	fo := NewFileObserver(ctx, pattern, fs.basedir, paths)

	doList := func() ([]string, error) {
		infoList, err := fs.ftpReadDir(patterndir)
		if ftpErrIsNotExist(err) {
			return nil, nil
		}
		if err != nil {
			log.Info().Err(err).Str("patterndir", patterndir).Msg("ftp: Error listing files")
			return nil, err
		}
		var fileList []string
		for _, info := range infoList {
			if !info.IsDir() {
				fileList = append(fileList, path.Join(patterndir, info.Name()))
			}
		}
		return fileList, nil
	}

	fileList, err := doList()
	if err != nil {
		return err
	}

	select {
	case paths <- "":
	case <-ctx.Done():
		return ctx.Err()
	}

	if err := fo.AddFiles(fileList...); err != nil {
		log.Err(err).Msg("ftp: Error filtering file list")
	}
	var errcount int
	for {
		var sleep time.Duration
		switch errcount {
		case 0:
			sleep = fs.PollingInterval
		case 1:
			sleep = 125 * time.Millisecond
		default:
			sleep *= 2
		}
		if sleep > fs.PollingInterval {
			sleep = fs.PollingInterval
		}
		log.Info().Dur("delay", sleep).Int("errcount", errcount).Msg("ftp: observe failed, will retry later")
		select {
		case <-time.After(sleep):
		case <-ctx.Done():
			return ctx.Err()
		}
		if fileList, err := doList(); err != nil {
			errcount++
		} else {
			if err := fo.AddFiles(fileList...); err != nil {
				log.Err(err).Msg("ftp: Error filtering file list")
				errcount++
			} else {
				errcount = 0
			}
		}
	}
}

// WaitForDeletion wath the given path every second until it does not exist
// anymore
func (fs *FtpFilesystem) WaitForDeletion(ctx context.Context, filepath string) error {
	filepath = path.Join(fs.basedir, filepath)
	for {
		switch fs.flavor {
		case FTPFlavorDefault:
			if _, err := fs.ftpStat(filepath); err != nil {
				if ftpErrIsNotExist(err) {
					return nil
				}
				return err
			}
		case FTPFlavorVSFTP:
			parent, name := path.Split(filepath)
			list, err := fs.ftpReadDir(parent)
			if err != nil {
				if ftpErrIsNotExist(err) {
					return nil
				}
				return err
			}
			var fileExists bool
			for _, info := range list {
				if info.Name() == name {
					fileExists = true

					break
				}
			}
			if !fileExists {
				return nil
			}
		}
		select {
		case <-time.After(fs.PollingInterval):
		case <-ctx.Done():
			return ctx.Err()
		}
	}
}

func init() {
	MustRegisterFilesystem("ftp", ftpFilesystemFactory)
	MustRegisterFilesystem("ftps", ftpFilesystemFactory)
}
