package filegateway

import (
	"io"
	"os"
)

type OnDiskBuffer struct {
	bufmaxsize int64
	dir        string
	pattern    string

	readCursor int

	buf      []byte
	tempFile *os.File
}

func NewOnDiskBuffer(bufmaxsize int64, dir, pattern string) *OnDiskBuffer {
	return &OnDiskBuffer{
		bufmaxsize: bufmaxsize,
		dir:        dir,
		pattern:    pattern,

		buf: make([]byte, 0, bufmaxsize/16),
	}
}

func (b *OnDiskBuffer) Read(data []byte) (int, error) {
	if b.tempFile != nil {
		return b.tempFile.Read(data)
	}
	if b.readCursor == len(b.buf) {
		return 0, io.EOF
	}
	n := copy(data, b.buf[b.readCursor:])
	b.readCursor += n
	return n, nil
}

func (b *OnDiskBuffer) Write(data []byte) (int, error) {
	if b.tempFile == nil && len(b.buf)+len(data) > int(b.bufmaxsize) {
		if err := b.switchToTempFile(); err != nil {
			return 0, err
		}
	}
	if b.tempFile != nil {
		return b.tempFile.Write(data)
	}
	b.buf = append(b.buf, data...)
	return len(data), nil
}

func (b *OnDiskBuffer) Seek(offset int64, whence int) (int64, error) {
	if b.tempFile != nil {
		return b.tempFile.Seek(offset, whence)
	}
	switch whence {
	case io.SeekStart:
		b.readCursor = int(offset)
	case io.SeekCurrent:
		b.readCursor += int(offset)
	case io.SeekEnd:
		b.readCursor = len(b.buf) + int(offset)
	}
	return int64(b.readCursor), nil
}

func (b *OnDiskBuffer) Reset() error {
	if b.tempFile != nil {
		if _, err := b.tempFile.Seek(0, io.SeekStart); err != nil {
			return err
		}
		if err := b.tempFile.Truncate(0); err != nil {
			return err
		}
	}
	b.buf = b.buf[0:0]
	b.readCursor = 0
	return nil
}

func (b *OnDiskBuffer) switchToTempFile() error {
	tempFile, err := os.CreateTemp(b.dir, b.pattern)
	if err != nil {
		return err
	}
	if _, err := tempFile.Write(b.buf); err != nil {
		_ = tempFile.Close()
		return err
	}
	b.tempFile = tempFile
	b.buf = nil
	return nil
}
